﻿B4A=true
B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=10
@EndOfDesignText@
#Event: ChatResponse(Response as string)
#Event: ImageResponse(Image as image)
#Event: TTSResponse(Folder as string, Filename as string)
#Event: Error(message as string)

Sub Class_Globals
	LogColor("Sub OpenAILibrary Class_Globals started", 0xFF210099)
	Private apiKey As String
	Private chatList As List
	Private mTarget As Object
	Private mEventname As String
	Public Model As String
	Public MODEL_GPT4 As Int = 2
	Public MODEL_GPT35_TURBO As Int = 1
	Public IMAGE_1_1 As Int = 1
	Public IMAGE_16_9 As Int = 2
	Public IMAGE_9_16 As Int = 3
	Public TTS_ALLOY As Int = 1
	Public TTS_ECHO As Int = 2
	Public TTS_FABLE As Int = 3
	Public TTS_ONYX As Int = 4
	Public TTS_NOVA As Int = 5
	Public TTS_SHIMMER As Int = 6
	Private TTS_VOICE As Int
	Private ImageAspectRatio As Int
	Private xui As XUI
	Private TTS_SPEED As Float
	Private apiUrl As String = "https://api.openai.com/v1/files"
	Private uploadApiUrl As String = "https://api.openai.com/v1/uploads"
	LogColor("Sub OpenAILibrary Class_Globals finished", 0xFF210099)
End Sub

'Initializes the object. You can add parameters to this method if needed.
Public Sub Initialize(Target As Object, EventName As String, key As String)
	LogColor("Sub OpenAILibrary Initialize started", 0xFF210099)
	apiKey = key
	mTarget = Target
	mEventname = EventName
	chatList.Initialize
	LogColor("Sub OpenAILibrary Initialize finished", 0xFF210099)
End Sub

'CHAT

public Sub SystemMessage(promt As String)
	LogColor("Sub OpenAILibrary SystemMessage started", 0xFF210099)
	Dim sanitizedPrompt As String = SanitizeInput(promt)
	Dim textChunks As List = SplitText(sanitizedPrompt)
    
	For Each chunk As String In textChunks
		chatList.Add(CreateMap("role": "system", "content": chunk))
	Next
    
	If chatList.Size > 0 Then
		Dim j As HttpJob
		j.Initialize("GetChatGPTResponse", Me)
		Dim JSONGenerator As JSONGenerator
		JSONGenerator.Initialize2(chatList)
		Log("JSON Chatlist: " & JSONGenerator.ToPrettyString(4))
		Dim chat_string As String
		chat_string = $"{"model":"${Model}","messages":${JSONGenerator.ToString}}"$
		j.PostString("https://api.openai.com/v1/chat/completions", chat_string)
		j.GetRequest.SetContentType("application/json")
		j.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
		j.GetRequest.SetHeader("OpenAI-Organization", "")
    
		' Log the request
		LogHttpRequest("GetChatGPTResponse", chat_string)
    
		Log("OpenAI: Generate Chat Response...")
	End If
	LogColor("Sub OpenAILibrary SystemMessage finished", 0xFF210099)
End Sub

Public Sub ResetChat
	LogColor("Sub OpenAILibrary ResetChat started", 0xFF210099)
	chatList.Clear
	LogColor("Sub OpenAILibrary ResetChat finished", 0xFF210099)
End Sub

Public Sub ChatMessage(promt As String)
	LogColor("Sub OpenAILibrary ChatMessage started", 0xFF210099)
	Dim sanitizedPrompt As String = SanitizeInput(promt)
	Dim textChunks As List = SplitText(sanitizedPrompt)
    
	For Each chunk As String In textChunks
		chatList.Add(CreateMap("role": "user", "content": chunk))
	Next
    
	If chatList.Size > 0 Then
		Dim j As HttpJob
		j.Initialize("GetChatGPTResponse", Me)
		Dim JSONGenerator As JSONGenerator
		JSONGenerator.Initialize2(chatList)
		Log("JSON Chatlist: " & JSONGenerator.ToPrettyString(4))
		Dim chat_string As String
		chat_string = $"{"model":"${Model}","messages":${JSONGenerator.ToString}}"$
		j.PostString("https://api.openai.com/v1/chat/completions", chat_string)
		j.GetRequest.SetContentType("application/json")
		j.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
		j.GetRequest.SetHeader("OpenAI-Organization", "")
    
		' Log the request
		LogHttpRequest("GetChatGPTResponse", chat_string)
    
		Log("OpenAI: Generate Chat Response...")
	End If
	LogColor("Sub OpenAILibrary ChatMessage finished", 0xFF210099)
End Sub

'TTS

Public Sub TextToSpeech(Prompt As String)
	LogColor("Sub OpenAILibrary TextToSpeech started", 0xFF210099)
	Dim sanitizedPrompt As String = SanitizeInput(Prompt)
	Dim textChunks As List = SplitText(sanitizedPrompt)
    
	If textChunks.Size = 0 Then
		Log("Error: Empty prompt provided for TTS")
		If xui.SubExists(mTarget, mEventname & "_Error", 1) Then
			CallSub2(mTarget, mEventname & "_Error", "Empty prompt provided for TTS")
		End If
		Return
	End If

	Dim config As Map = CreateMap("model": "tts-1-hd", "voice": getVoice, "speed": TTS_SPEED, "apiKey": apiKey)
	FetchAudio(textChunks, config)
	LogColor("Sub OpenAILibrary TextToSpeech finished", 0xFF210099)
End Sub

Private Sub FetchAudio(textChunks As List, config As Map)
	LogColor("Sub OpenAILibrary FetchAudio started", 0xFF210099)
	Dim rpm As Int = 100 ' Maximum requests per minute
	Dim interval As Int = 60000 / rpm ' Time between requests in milliseconds
    
	For i = 0 To textChunks.Size - 1
		Dim chunk As String = textChunks.Get(i)
        
		If i > 0 And i Mod rpm = 0 Then
			Sleep(60000) ' Wait for a minute after 100 requests
		Else If i > 0 Then
			Sleep(interval) ' Wait the required interval before the next request
		End If
        
		Dim j As HttpJob
		j.Initialize("ttsJob", Me)
		Dim request As String = $"{"model":"${config.Get("model")}","input":"${chunk}","voice":"${config.Get("voice")}","speed":${config.Get("speed")}}"$
		j.PostString("https://api.openai.com/v1/audio/speech", request)
		j.GetRequest.SetContentType("application/json")
		j.GetRequest.SetHeader("Authorization", "Bearer " & config.Get("apiKey"))
		
		' Log the request
		LogHttpRequest("ttsJob", request)
	Next
	LogColor("Sub OpenAILibrary FetchAudio finished", 0xFF210099)
End Sub

Public Sub setTTSSpeed(speed As Float)
	LogColor("Sub OpenAILibrary setTTSSpeed started", 0xFF210099)
	TTS_SPEED = speed
	LogColor("Sub OpenAILibrary setTTSSpeed finished", 0xFF210099)
End Sub

Public Sub setTTSVoice(voice As Int)
	LogColor("Sub OpenAILibrary setTTSVoice started", 0xFF210099)
	TTS_VOICE=voice
	LogColor("Sub OpenAILibrary setTTSVoice finished", 0xFF210099)
End Sub

private Sub getVoice As String
	LogColor("Sub OpenAILibrary getVoice started", 0xFF210099)
	Select TTS_VOICE
		Case TTS_ALLOY
			Return "alloy"
		Case TTS_ECHO
			Return "echo"
		Case TTS_FABLE
			Return "fable"
		Case TTS_ONYX
			Return "onyx"
		Case TTS_NOVA
			Return "nova"
		Case TTS_SHIMMER
			Return "shimmer"
	End Select
	LogColor("Sub OpenAILibrary getVoice finished", 0xFF210099)
	Return "alloy"
End Sub

'IMAGE

Public Sub generateImage(Prompt As String)
	LogColor("Sub OpenAILibrary generateImage started", 0xFF210099)
	Dim sanitizedPrompt As String = SanitizeInput(Prompt)
	Dim textChunks As List = SplitText(sanitizedPrompt)
    
	If textChunks.Size = 0 Then
		Log("Error: Empty prompt provided for image generation")
		Return
	End If
    
	' Combine all chunks into a single prompt, separated by spaces
	Dim fullPrompt As StringBuilder
	fullPrompt.Initialize
	For Each chunk As String In textChunks
		If fullPrompt.Length > 0 Then
			fullPrompt.Append(" ")
		End If
		fullPrompt.Append(chunk)
	Next
    
	' Truncate if exceeds maximum length (e.g., 1000 characters)
	Dim maxPromptLength As Int = 1000
	If fullPrompt.Length > maxPromptLength Then
		fullPrompt.Remove(maxPromptLength, fullPrompt.Length)
		Log("Warning: Prompt truncated to " & maxPromptLength & " characters for image generation")
	End If
    
	Dim j As HttpJob
	j.Initialize("generateimage", Me)
	Dim request As String = $"{"model":"dall-e-3","prompt":"${fullPrompt.ToString}","n":1,"size":"${getAspectRatio}"}"$
	j.PostString("https://api.openai.com/v1/images/generations", request)
	j.GetRequest.SetContentType("application/json")
	j.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	j.GetRequest.SetHeader("OpenAI-Organization", "")
	' Log the request
	LogHttpRequest("generateimage", request)
	Log("OpenAI: Generate Image...")
	LogColor("Sub OpenAILibrary generateImage finished", 0xFF210099)
End Sub

Public Sub setImageAspectRatio(ratio As Int)
	LogColor("Sub OpenAILibrary setImageAspectRatio started", 0xFF210099)
	ImageAspectRatio = ratio
	LogColor("Sub OpenAILibrary setImageAspectRatio finished", 0xFF210099)
End Sub

private Sub getAspectRatio As String
	LogColor("Sub OpenAILibrary getAspectRatio started", 0xFF210099)
	Select ImageAspectRatio
		Case IMAGE_1_1
			Return "1024x1024"
		Case IMAGE_16_9
			Return "1792x1024"
		Case IMAGE_9_16
			Return "1024x1792"
	End Select
	LogColor("Sub OpenAILibrary getAspectRatio finished", 0xFF210099)
	Return "1024x1024"
End Sub

'UTILITIES FOR CHAT,TTS,IMAGE

Private Sub SanitizeInput(input As String) As String
	LogColor("Sub OpenAILibrary SanitizeInput started", 0xFF210099)
	LogColor("Sub OpenAILibrary SanitizeInput finished", 0xFF210099)
	Return input.Replace(CRLF, " ").Replace(Chr(34), "'")
End Sub

Private Sub SplitText(text As String) As List
	LogColor("Sub OpenAILibrary SplitText started", 0xFF210099)
	Dim chunks As List
	chunks.Initialize
    
	Dim maxChunkSize As Int = 4096
	Dim delimiters As List
	delimiters.Initialize
	delimiters.AddAll(Array As String(". ", "? ", "! ", CRLF))
    
	Do While text.Length > 0
		If text.Length <= maxChunkSize Then
			chunks.Add(text)
			Exit
		End If
        
		Dim end1 As Int = maxChunkSize
		For Each delimiter As String In delimiters
			Dim pos As Int = text.SubString2(0, maxChunkSize).LastIndexOf(delimiter)
			If pos > -1 Then
				end1 = pos + delimiter.Length
				Exit
			End If
		Next
        
		If end1 = maxChunkSize Then
			' If no delimiter was found, force a split at maxChunkSize
			end1 = maxChunkSize
		End If
        
		chunks.Add(text.SubString2(0, end1))
		text = text.SubString(end1)
	Loop
	LogColor("Sub OpenAILibrary SplitText finished", 0xFF210099)
	Return chunks
End Sub

private Sub ParseJson(json As String) As String
	LogColor("Sub OpenAILibrary ParseJson started", 0xFF210099)
	Dim parser As JSONParser
	parser.Initialize(json)
	Dim Root1 As Map
	Root1 = parser.NextObject
	Dim choices As List
	choices = Root1.Get("choices")
	Dim choiceIndex As Int
	Dim content As String
	For choiceIndex = 0 To choices.Size - 1
		Dim choice As Map
		choice = choices.Get(choiceIndex)
		Dim message As Map
		message = choice.Get("message")
		If content <> "" Then content = content & CRLF
		content = content & message.Get("content")
		chatList.Add(CreateMap("role" : "assistant", "content" : content))
	Next
	LogColor("Sub OpenAILibrary ParseJson finished", 0xFF210099)
	Return content
End Sub

'OPENAI FILE LIST TABLE

Public Sub RetrieveFile(fileId As String)
	LogColor("Sub OpenAILibrary RetrieveFile started", 0xFF210099)
	'***This sub is not tested because I can't download a file that is uploaded with purpose assistants.
	Dim job As HttpJob
	job.Initialize("retrieveFileJob", Me)
	job.Download(apiUrl & "/" & fileId & "/content")
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	LogHttpRequest("retrieveFileJob", apiUrl & "/" & fileId & "/content") ' Log the request.
	
	Wait For (job) JobDone (job As HttpJob)
    
	If job.Success Then
		Dim parser As JSONParser
		parser.Initialize(job.GetString)
	
		' Create a Map from the JSON object
		Dim resultMap As Map = parser.NextObject()
		Dim FileName As String = resultMap.Get("filename")
    
		' Modify the "object" field
		resultMap.Put("object", "retreivefile") ' Change from "file" to "uploadfile"

		' Reassemble the Map to JSON
		Dim jsonGen As JSONGenerator
		jsonGen.Initialize(resultMap)
    
		Dim newJson As String = jsonGen.ToString ' Create the new JSON string

		' Log the newly created JSON
		Log("Modified JSON: " & newJson)
		If xui.SubExists(mTarget, mEventname & "_Response", 2) Then
			CallSub2(mTarget, mEventname & "_Response",newJson)
		Else
			xui.MsgboxAsync("There is no Sub OpenAI_Response(response As String) in the user's code.","ERROR")
		End If
				
		Dim inputStream As InputStream = job.GetInputStream
		If xui.SubExists(mTarget, mEventname & "_RetreiveFileResponse", 1) Then
			CallSub2(mTarget, mEventname & "_RetreiveFileResponse", inputStream)
		Else
			Log("OpenAI_RetreiveFileResponse(inputStream As InputStream) sub is missing from user code.")
		End If
	End If
	
	Dim j As HttpJob
	j.Initialize("DownloadRetrievedFile", Me)
	LogHttpRequest("DownloadRetrievedFile", FileName) ' Log the request
	j.Download(FileName)
	LogColor("Sub OpenAILibrary RetrieveFile finished", 0xFF210099)
End Sub

Public Sub ListFiles
	LogColor("Sub OpenAILibrary ListFiles started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("listFilesJob", Me)
	job.Download(apiUrl)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	LogHttpRequest("listFilesJob", apiUrl) ' Log the request.
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	job.GetRequest.SetHeader("limit", "100")
	LogColor("Sub OpenAILibrary ListFiles finished", 0xFF210099)
End Sub

Public Sub DeleteFile(fileId As String)
	LogColor("Sub OpenAILibrary DeleteFile started", 0xFF210099)
	Log("Here at DeleteFile")
    Dim job As HttpJob
    job.Initialize("deleteFileJob", Me)
    job.Tag = fileId ' Store the fileId in the job's tag
    job.Delete(apiUrl & "/" & fileId)
    job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogHttpRequest("deleteFileJob", apiUrl & "/" & fileId) ' Log the request.
	LogColor("Sub OpenAILibrary DeleteFile finished", 0xFF210099)
End Sub

'ASSISTANTS

Public Sub CreateAssistant(assistantName As String, assistantInstructions As String, assistantModel As String, assistantTools As List)
	LogColor("Sub OpenAILibrary CreateAssistant started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("CreateAssistant", Me)
    
	' Create a map for the JSON body
	Dim jsonMap As Map
	jsonMap.Initialize
	jsonMap.Put("instructions", assistantInstructions)
	jsonMap.Put("name", assistantName)
	jsonMap.Put("tools", assistantTools) ' Ensure this is a list of maps
	jsonMap.Put("model", assistantModel)
    
	' Convert the map to a JSON string
	Dim json As JSONGenerator
	json.Initialize(jsonMap)
	Dim jsonString As String = json.ToString
    
	' Post the JSON string directly
	job.PostString("https://api.openai.com/v1/assistants", jsonString)
    
	' Set headers
	job.GetRequest.SetContentType("application/json")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary CreateAssistant finished", 0xFF210099)
End Sub

Public Sub ListAssistants(limit As String, order As String, after As String, before As String)
    LogColor("Sub OpenAILibrary ListAssistants started", 0xFF210099)
    
    Dim job As HttpJob
    job.Initialize("ListAssistants", Me)
    
    ' Start with the base URL
    Dim url As String = "https://api.openai.com/v1/assistants"
    
    Dim parameters As String = ""
    
    ' Add the limit parameter if it's not blank
    If limit <> "" Then
        parameters = parameters & $"limit=${limit}"$
    End If
    
    ' Add the order parameter if it's not blank
    If order <> "" Then
        If parameters <> "" Then parameters = parameters & "&"
        parameters = parameters & $"order=${order}"$
    End If
    
    ' Add the after parameter if it's not blank
    If after <> "" Then
        If parameters <> "" Then parameters = parameters & "&"
        parameters = parameters & $"after=${after}"$
    End If
    
    ' Add the before parameter if it's not blank
    If before <> "" Then
        If parameters <> "" Then parameters = parameters & "&"
        parameters = parameters & $"before=${before}"$
    End If
    
    ' If parameters are present, append them to the URL
    If parameters <> "" Then
        url = url & "?" & parameters
    End If

    job.Download(url)
    job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
    job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
    ' Log the request
    LogHttpRequest("ListAssistants", url)
    LogColor("Sub OpenAILibrary ListAssistants finished", 0xFF210099)
End Sub

Public Sub DeleteAssistant(AssistantID As String)
	LogColor("Sub OpenAILibrary DeleteAssistant started", 0xFF210099)
	Private Job1 As HttpJob
	' Initialize the HTTP job
	Job1.Initialize("DeleteAssistant", Me)
    
	' Set the DELETE request URL
	Dim url As String = $"https://api.openai.com/v1/assistants/${AssistantID}"$
	Job1.Delete(url) ' Send a DELETE request to the specified URL

	' Set the necessary headers
	Job1.GetRequest.SetHeader("Content-Type", "application/json")
	Job1.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	Job1.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary DeleteAssistant finished", 0xFF210099)
End Sub

Public Sub EditAssistant(AssistantID As String)
	LogColor("Sub OpenAILibrary EditAssistant started", 0xFF210099)
	'*** Add code for this.
	LogColor("Sub OpenAILibrary EditAssistant finished", 0xFF210099)
End Sub

Public Sub RetrieveAssistant(AssistantID As String) 'This is really activating an assistant that already exists, not retrieving it.
	LogColor("Sub OpenAILibrary RetrieveAssistant started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("RetrieveAssistant", Me)
	'https://api.openai.com/v1/assistants/{assistant_id}
	job.PostString("https://api.openai.com/v1/assistants/" & AssistantID,"")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("Content-Type", "application/json")
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary RetrieveAssistant finished", 0xFF210099)
End Sub

Public Sub AddVectorStoreToAssistant(assistantId As String, vectorStoreId As String)
	LogColor("Sub OpenAILibrary UpdateAssistantWithVectorStore started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("UpdateAssistant", Me)
	Dim json As JSONGenerator
	json.Initialize(CreateMap("tool_resources": CreateMap("file_search": CreateMap("vector_store_ids": Array As String(vectorStoreId)))))
	Dim request As String = json.ToString
	job.PostString("https://api.openai.com/v1/assistants/" & assistantId, request)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("Content-Type", "application/json")
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("UpdateAssistant", request)
	LogColor("Sub OpenAILibrary UpdateAssistantWithVectorStore finished", 0xFF210099)
End Sub

'Runs
Public Sub CreateRun(assistantId As String, threadID As String)
	LogColor("Sub OpenAILibrary CreateRun started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("CreateRun", Me) 'Object response is "thread.run"

	Dim parameters As String
	parameters = $"{"assistant_id":"${assistantId}"}"$
	job.PostString("https://api.openai.com/v1/threads/" & threadID & "/runs", parameters)
	job.GetRequest.SetContentType("application/json")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Organization", "")
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary CreateRun finished", 0xFF210099)
End Sub

Public Sub ListRuns(CurrentThreadID As String, limit As String, after As String, before As String, order As String)
	LogColor("Sub OpenAILibrary ListRuns started", 0xFF210099)
    
	' Start with the base URL
	Dim baseUrl As String = $"https://api.openai.com/v1/threads/${CurrentThreadID}/runs"$

	Dim url As String = baseUrl
	Dim parameters As String = ""

	' Add the before parameter if it's not blank
	If before <> "" Then
		If parameters <> "" Then parameters = parameters & "&" ' Add '&' if parameters are already present
		parameters = parameters & $"before=${before}"$
	End If
    
	' Add the after parameter if it's not blank
	If after <> "" Then
		If parameters <> "" Then parameters = parameters & "&" ' Add '&' if parameters are already present
		parameters = parameters & $"after=${after}"$
	End If
    
	' Add the limit parameter if it's not blank
	If limit <> "" Then
		If parameters <> "" Then parameters = parameters & "&" ' Add '&' if parameters are already present
		parameters = parameters & $"limit=${limit}"$
	End If

	' Add the order parameter if it's not blank
	If order <> "" Then
		If parameters <> "" Then parameters = parameters & "&" ' Add '&' if parameters are already present
		parameters = parameters & $"order=${order}"$  ' Assuming order is the string parameter you want to add
	End If

	' Check if we have any parameters to append
	If parameters <> "" Then
		url = url & "?" & parameters ' Append parameters to the URL
	End If

	Dim job As HttpJob
	job.Initialize("ListRuns", Me)
	job.Download(url)  ' Download the constructed URL
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("ListRuns", url)
	LogColor("Sub OpenAILibrary ListRuns finished", 0xFF210099)
End Sub

Public Sub RunAssistant(assistantId As String, threadId As String)
	LogColor("Sub OpenAILibrary RunAssistant started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("RunAssistant", Me)
	Dim json As JSONGenerator
	json.Initialize(CreateMap("assistant_id": assistantId))
	Dim request As String = json.ToString
	job.PostString($"https://api.openai.com/v1/threads/${threadId}/runs"$, request)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("Content-Type", "application/json")
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("RunAssistant", request)
	LogColor("Sub OpenAILibrary RunAssistant finished", 0xFF210099)
End Sub

Public Sub CheckRunStatus(threadId As String, runId As String)
	LogColor("Sub OpenAILibrary CheckRunStatus started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("CheckRunStatus", Me)
	Dim url As String = $"https://api.openai.com/v1/threads/${threadId}/runs/${runId}"$
	job.Download(url)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("CheckRunStatus", url)
	LogColor("Sub OpenAILibrary CheckRunStatus finished", 0xFF210099)
End Sub

Public Sub RetrieveRun(threadId As String, runId As String)
	LogColor("Sub OpenAILibrary CheckRunStatus started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("RetrieveRun", Me)
	job.Download("https://api.openai.com/v1/threads/" & threadId & "/runs/" & runId)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("CheckRunStatus", "https://api.openai.com/v1/threads/" & threadId & "/runs/" & runId)
	LogColor("Sub OpenAILibrary ListRuns finished", 0xFF210099)
End Sub

Public Sub CancelRun(threadId As String, runId As String)
	LogColor("Sub OpenAILibrary CancelRun started", 0xFF210099)
    
	Dim job As HttpJob
	job.Initialize("CancelRun", Me)

	' Construct the URL for the cancel operation
	Dim url As String = $"https://api.openai.com/v1/threads/${threadId}/runs/${runId}/cancel"$

	job.PostString(url, "")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	job.GetRequest.SetHeader("Content-Type", "application/json")

	' Log the request
	LogHttpRequest("CancelRun", url)

	LogColor("Sub OpenAILibrary CancelRun finished", 0xFF210099)
End Sub

'Threads
Public Sub CreateThread()
	LogColor("Sub OpenAILibrary CreateThread started", 0xFF210099)
	Dim Job As HttpJob
	Job.Initialize("CreateThread", Me)
    
	Job.PostString("https://api.openai.com/v1/threads", "{}")
	Job.GetRequest.SetContentType("application/json")
	Job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	Job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary CreateThread finished", 0xFF210099)
End Sub

Public Sub ListThreads  'Can't list Threads due to OpenAI not supporting it. Spreadsheet is used.
	LogColor("Sub OpenAILibrary ListThreads started", 0xFF210099)
	'Listing threads is not allowed by the API at this time.  See my text file about this in the B4J folder.
	'So, this commented code can't be used until this is fixed and an endpoint in the API is established for it.
'	Dim job As HttpJob
'	job.Initialize("ListThreads", Me)
'	job.Download("https://api.openai.com/v1/threads")
'	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
'	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
'	LogHttpRequest("ListThreads", "https://api.openai.com/v1/threads") ' Log the request.
	LogColor("Sub OpenAILibrary ListThreads finished", 0xFF210099)
End Sub

Public Sub AddMessageToThread(threadId As String, role As String, content As String)
	LogColor("Sub OpenAILibrary AddMessageToThread started", 0xFF210099)
	
	threadId = "thread_lookCzck2zvlisp6kdYDgMCg"
	content = "This is a test message"
	role = "user"
	
	Dim Job As HttpJob
	Job.Initialize("CreateThreadMessageJob", Me)
    
	' Construct the JSON request body
	Dim json As String = $"{
        "role": "${role}",
        "content": "${content}",
        "attachments": null,
        "metadata": {}
    }"$

	' Set the API endpoint for creating a message in the thread
	Dim createMessageApiUrl As String = $"https://api.openai.com/v1/threads/${threadId}/messages"$

	' Send the POST request
	Job.PostString(createMessageApiUrl, json)
    
	' Set the request headers
	Job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	Job.GetRequest.SetContentType("application/json")
	Job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary AddMessageToThread finished", 0xFF210099)
End Sub

Public Sub DeleteThread(threadId As String)
	LogColor("Sub OpenAILibrary DeleteThread started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("DeleteThread", Me)
    
	' Set the API endpoint for deleting a thread
	Dim deleteThreadApiUrl As String = $"https://api.openai.com/v1/threads/${threadId}"$

	' Send the DELETE request
	job.Delete(deleteThreadApiUrl)
    
	' Set the request headers
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")

	' Log the request
	LogHttpRequest("DeleteThread", deleteThreadApiUrl)
	'*** Add to the code to allow the user to select where to store the LogHttpRequest
	LogColor("Sub OpenAILibrary DeleteThread finished", 0xFF210099)
End Sub

'Messages

Public Sub GetMessagesFromThread(thread_id As String, limit As String, order As String, after As String, before As String, run_id As String)
	LogColor("Sub OpenAILibrary GetMessagesFromThread started", 0xFF210099)
    
	Dim job As HttpJob
	job.Initialize("GetMessagesFromThread", Me)
    
	' Start with the base URL
	Dim baseUrl As String = $"https://api.openai.com/v1/threads/${thread_id}/messages"$

	Dim url As String = baseUrl
	Dim parameters As String = ""

	' Add the limit parameter if it's not blank
	If limit <> "" Then
		parameters = parameters & $"limit=${limit}"$
	End If

	' Add the order parameter if it's not blank
	If order <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"order=${order}"$
	End If

	' Add the after parameter if it's not blank
	If after <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"after=${after}"$
	End If

	' Add the before parameter if it's not blank
	If before <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"before=${before}"$
	End If

	' Add the run_id parameter if it's not blank
	If run_id <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"run_id=${run_id}"$
	End If

	' If parameters are present, append them to the URL
	If parameters <> "" Then
		url = url & "?" & parameters
	End If

	job.Download(url)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("GetMessagesFromThread", url)
	LogColor("Sub OpenAILibrary GetMessagesFromThread finished", 0xFF210099)
End Sub

Public Sub DeleteMessageFromThread(threadId As String, messageId As String)
	LogColor("Sub OpenAILibrary DeleteMessageFromThread started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("DeleteMessage", Me)

	' Set the URL for the API call
	Dim url As String = $"https://api.openai.com/v1/threads/${threadId}/messages/${messageId}"$

	' Set the request headers
	job.GetRequest.SetHeader("Content-Type", "application/json")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
    job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")

	' Send the DELETE request
	job.Delete(url)
	LogColor("Sub OpenAILibrary DeleteMessageFromThread finished", 0xFF210099)
End Sub

' Subroutine to retrieve a specific message from a thread
Public Sub RetrieveMessageFromThread(threadId As String, messageId As String)
	LogColor("Sub OpenAILibrary RetrieveMessageFromThread started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("GetMessage", Me)
	job.Download("https://api.openai.com/v1/threads/" & threadId & "/messages/" & messageId)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary RetrieveMessageFromThread finished", 0xFF210099)
End Sub

'UPLOAD FILES

Public Sub UploadFile(filePath As String, purpose As String)
	LogColor("Sub OpenAILibrary UploadFile started", 0xFF210099)
	Log("Here Upload File")
    Dim job As HttpJob
    job.Initialize("UploadFile", Me)
    job.Tag = filePath
	
    Dim mfd As MultipartFileData
    mfd.Initialize
    mfd.Dir = File.GetFileParent(filePath)
    mfd.FileName = File.GetName(filePath)
    mfd.KeyName = "file"
    mfd.ContentType = GetMimeType(mfd.FileName)
    
    ' Create the map for purpose
    Dim purposeMap As Map
    purposeMap.Initialize
    purposeMap.Put("purpose", purpose)
    
    ' Convert file data to a string for logging
    Dim fileDataString As String
    fileDataString = $"Dir: ${mfd.Dir}, FileName: ${mfd.FileName}, KeyName: ${mfd.KeyName}, ContentType: ${mfd.ContentType}"$
    
    ' Log the request including purpose and file data
    Dim requestString As String = $"Purpose: ${purpose}, FileData: ${fileDataString}"$
    LogHttpRequest("UploadFile", requestString)
    
    ' Post the multipart request
    job.PostMultipart("https://api.openai.com/v1/files", purposeMap, Array(mfd))
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary UploadFile finished", 0xFF210099)
End Sub

Public Sub InitializeUpload(Purpose As String, Folder As String, FileName As String, OpenAIapiKey As String)
	LogColor("Sub OpenAILibrary InitializeUpload started", 0xFF210099)
	Log("Folder = " & Folder)
	Log("FileName = " & FileName)
	Private uploadApiUrl As String = "https://api.openai.com/v1/uploads"
	'Dim FilePath As String = Folder & "\" & FileName
	Dim fileSize As Long = File.Size(Folder, FileName)
	Dim mimeType As String = GetMimeType(FileName) ' Adjust MIME type as needed

	' Initialize the upload
	Dim Job As HttpJob
	Job.Initialize("initializeUploadJob", Me)
	Dim json As String = $"{"filename": "${FileName}", "purpose": "${Purpose}", "bytes": ${fileSize}, "mime_type": "${mimeType}"}"$
	Job.PostString(uploadApiUrl, json)
	Job.GetRequest.SetHeader("Authorization", "Bearer " & OpenAIapiKey)
	Job.GetRequest.SetContentType("application/json")
	Job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	Wait For (Job) JobDone (Job As HttpJob)
    
	If Job.Success Then
		Log("Initialize Upload Job Response = " & Job.GetString)
		Dim parser As JSONParser
		parser.Initialize(Job.GetString)
		Dim result As Map = parser.NextObject
		Dim FileID As String = result.Get("id")
		Log("File ID = " & FileID)

		'Reply to the user code
		' Modify the "object" field
		result.Put("object", "initializeupload") ' Change from "file" to "uploadfile"
		' Reassemble the Map to JSON
		Dim jsonGen As JSONGenerator
		jsonGen.Initialize(result)
		Dim newJson As String = jsonGen.ToString ' Create the new JSON string
		' Log the newly created JSON
		Log("Modified JSON: " & newJson)
		If xui.SubExists(mTarget, mEventname & "_Response", 2) Then
			CallSub2(mTarget, mEventname & "_Response",newJson)
		Else
			xui.MsgboxAsync("There is no Sub OpenAI_Response(response As String) in the user's code.","ERROR")
		End If

		' Proceed to upload parts
		Dim partSize As Long = 64 * 1024 * 1024 ' 64 MB
		Dim partCount As Int = Ceil(fileSize / partSize)
		Log("PartCount = " & partCount)
		Dim partIds As List
		partIds.Initialize
        Job.Release
		
		For i = 0 To partCount - 1
			Dim start As Long = i * partSize
			Dim end1 As Long = Min((i + 1) * partSize, fileSize)
			Dim length As Long = end1 - start
			Log("Uploading part " & (i + 1) & " of " & partCount & ", bytes " & start & " to " & end1)
            
			' Read the part data
			Dim raf As RandomAccessFile
			raf.Initialize(Folder, FileName, False)
			Dim partData(length) As Byte
			raf.ReadBytes(partData, 0, length, start)
			raf.Close
            
			' Build the multipart form-data body
			Dim stream As OutputStream
			stream.InitializeToBytesArray(0)
            
			' Define boundary
			Dim boundary As String = "---------------------------1461124740692"
			Dim eol As String = Chr(13) & Chr(10)
            
			' Build the header
			Dim fileHeader As String = _
$"--${boundary}
Content-Disposition: form-data; name="data"; filename="${FileName}"
Content-Type: ${mimeType}

"$
			fileHeader = fileHeader.Replace(CRLF, eol)
			Dim headerBytes() As Byte = fileHeader.GetBytes("UTF8")
			stream.WriteBytes(headerBytes, 0, headerBytes.Length)
            
			' Write the data
			stream.WriteBytes(partData, 0, partData.Length)
            
			' Write the end boundary
			Dim endBoundary As String = eol & "--" & boundary & "--" & eol
			Dim endBoundaryBytes() As Byte = endBoundary.GetBytes("UTF8")
			stream.WriteBytes(endBoundaryBytes, 0, endBoundaryBytes.Length)
            
			' Initialize and send the job
			Dim partJob As HttpJob
			Dim ix As Int = i + 1
			partJob.Tag = "File # "& ix & " of " & partCount
			partJob.Initialize("uploadPartJob", Me)
			Dim url As String = "https://api.openai.com/v1/uploads/" & FileID & "/parts"
			partJob.PostBytes(url, stream.ToBytesArray)
			partJob.GetRequest.SetHeader("Authorization", "Bearer " & OpenAIapiKey)
			partJob.GetRequest.SetContentType("multipart/form-data; boundary=" & boundary)

			' Log the Request Details
			Log("Uploading to URL: " & url)
			Log("Authorization: Bearer " & OpenAIapiKey)
			Log("Content-Type: multipart/form-data; boundary=" & boundary)
			' Log partial stream content, first 256 characters, for inspection
			'Dim streamContent As String = BytesToString(stream.ToBytesArray, 0, Min(256, stream.ToBytesArray.Length), "UTF8")
			'Log("Request Body Content (partial): " & streamContent)
            
			Wait For (partJob) JobDone (partJob As HttpJob)
            
			If partJob.Success Then
				' Parse the response
				Dim parser2 As JSONParser
				parser2.Initialize(partJob.GetString)
				Dim result2 As Map = parser2.NextObject
				Dim partId As String = result2.Get("id")
				
				'Report back to the user's code
				If xui.SubExists(mTarget, mEventname & "_Response", 2) Then
					CallSub2(mTarget, mEventname & "_Response",partJob.GetString)
				Else
					xui.MsgboxAsync("There is no Sub OpenAI_Response(response As String) in the user's code.","ERROR")
				End If
				partIds.Add(partId)
				Log("Uploaded part " & (i + 1) & " with id " & partId)
			Else
				Log("Error uploading part " & (i + 1) & ": " & partJob.ErrorMessage)
				' Handle the error appropriately
				Return ' Exit the sub if there's an error
			End If
            
			partJob.Release
		Next
        
		' Complete the upload
		Dim partIdsJson As String = "["
		For i = 0 To partIds.Size - 1
			Dim partId As String = partIds.Get(i)
			partIdsJson = partIdsJson & $" "${partId}" "$
			If i < partIds.Size - 1 Then
				partIdsJson = partIdsJson & ","
			End If
		Next
		partIdsJson = partIdsJson & "]"
        
		' Build the complete request JSON
		Dim completeJson As String = $"{"part_ids": ${partIdsJson}}"$

		' Send the complete request
		Dim completeJob As HttpJob
		completeJob.Initialize("completeUploadJob", Me)
		Dim completeUrl As String = "https://api.openai.com/v1/uploads/" & FileID & "/complete"
		completeJob.PostString(completeUrl, completeJson)
		completeJob.GetRequest.SetHeader("Authorization", "Bearer " & OpenAIapiKey)
		completeJob.GetRequest.SetContentType("application/json")

		Wait For (completeJob) JobDone (completeJob As HttpJob)

		If completeJob.Success Then
			Log("Upload completed successfully.")
			Log("Response: " & completeJob.GetString)
			
			'Report back to the user's code
			If xui.SubExists(mTarget, mEventname & "_Response", 2) Then
				CallSub2(mTarget, mEventname & "_Response",completeJob.GetString)
			Else
				xui.MsgboxAsync("There is no Sub OpenAI_Response(response As String) in the user's code.","ERROR")
			End If
		Else
			Log("Error completing upload: " & completeJob.ErrorMessage)
		End If

		completeJob.Release
	Else
		Log("Error initializing upload: " & Job.ErrorMessage)
	End If
	Job.Release
	LogColor("Sub OpenAILibrary InitializeUpload finished", 0xFF210099)
End Sub

Sub CancelUpload(uploadId As String)
	LogColor("Sub OpenAILibrary CancelUpload started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("CancelUpload", Me)
	job.PostString(uploadApiUrl & "/" & uploadId & "/cancel", "")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	
	LogHttpRequest("CancelUpload", uploadApiUrl & "/" & uploadId & "/cancel") ' Log the request.
	LogColor("Sub OpenAILibrary CancelUpload finished", 0xFF210099)
End Sub

Public Sub CreateVectorStore(VectorStoreName As String)
	LogColor("Sub OpenAILibrary CreateVectorStore started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("CreateVectorStore", Me)

	Dim parameters As String
	parameters = $"{"name":"${VectorStoreName}"}"$
	job.PostString("https://api.openai.com/v1/vector_stores/", parameters)
	job.GetRequest.SetContentType("application/json")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Organization", "")
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")

	' Log the request
	LogHttpRequest("https://api.openai.com/v1/vector_stores/", parameters)
	LogColor("Sub OpenAILibrary CreateVectorStore finished", 0xFF210099)
End Sub

Private Sub GetMimeType(fileName As String) As String
	LogColor("Sub OpenAI Library GetMimeType started", 0xFF009901)
	'Supported File Types:
	'https://platform.openai.com/docs/assistants/tools/code-interpreter/supported-files
	Dim extension As String = fileName.SubString(fileName.LastIndexOf(".") + 1).ToLowerCase
	LogColor("Sub OpenAI Library GetMimeType finished", 0xFFFF7100)
	Select extension
		Case "txt"
			Return "text/plain"
		Case "html", "htm"
			Return "text/html"
		Case "jpg", "jpeg"
			Return "image/jpeg"
		Case "png"
			Return "image/png"
		Case "pdf"
			Return "application/pdf"
		Case "doc"
			Return "application/msword"
		Case "docx"
			Return "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
		Case "xls"
			Return "application/vnd.ms-excel"
		Case "xlsx"
			Return "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
		Case "c"
			Return "text/x-c"
		Case "cs"
			Return "text/x-csharp"
		Case "cpp"
			Return "text/x-c++"
		Case "java"
			Return "text/x-java"
		Case "json"
			Return "application/json"
		Case "md"
			Return "text/markdown"
		Case "php"
			Return "text/x-php"
		Case "pptx"
			Return "application/vnd.openxmlformats-officedocument.presentationml.presentation"
		Case "py"
			Return "text/x-python"
		Case "rb"
			Return "text/x-ruby"
		Case "tex"
			Return "text/x-tex"
		Case "css"
			Return "text/css"
		Case "js"
			Return "text/javascript"
		Case "sh"
			Return "application/x-sh"
		Case "ts"
			Return "application/typescript"
		Case "csv"
			Return "application/csv"
		Case "gif"
			Return "image/gif"
		Case "pkl"
			Return "application/octet-stream"
		Case "tar"
			Return "application/x-tar"
		Case "xml"
			Return "text/xml"
		Case "zip"
			Return "application/zip"
		Case Else
			Return "application/octet-stream" ' Default MIME type for unknown extensions
	End Select
End Sub

'VECTOR STORES

Public Sub ListVectorStores(limit As String, order As String, after As String, before As String, filter As String)
	LogColor("Sub OpenAILibrary ListVectorStores started", 0xFF210099)
    
	Dim job As HttpJob
	job.Initialize("ListVectorStores", Me)
    
	Dim url As String = "https://api.openai.com/v1/vector_stores"
	Dim parameters As String = ""

	' Add the limit parameter if it's not blank and is a valid number
	If limit <> "" Then
		parameters = parameters & $"limit=${limit}"$
	End If

	' Add the order parameter if it's not blank
	If order <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"order=${order}"$
	End If

	' Add the after parameter if it's not blank
	If after <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"after=${after}"$
	End If

	' Add the before parameter if it's not blank
	If before <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"before=${before}"$
	End If

	' Add the filter parameter if it's not blank
	If filter <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"filter=${filter}"$
	End If

	' If parameters are present, append them to the URL
	If parameters <> "" Then
		url = url & "?" & parameters
	End If

	job.Download(url)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("ListVectorStores", url)
	LogColor("Sub OpenAILibrary ListVectorStores finished", 0xFF210099)
End Sub

Public Sub AddFileToVectorStore(vectorStoreId As String, fileId As String)
	LogColor("Sub OpenAILibrary Class_Globals started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("AddFileToVectorStore", Me)
	Dim json As JSONGenerator
	json.Initialize(CreateMap("file_ids": Array(fileId)))
	Dim request As String = json.ToString
	job.PostString("https://api.openai.com/v1/vector_stores/" & vectorStoreId & "/file_batches", request)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("Content-Type", "application/json")
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("AddFileToVectorStore", request)
	LogColor("Sub OpenAILibrary CreateVectorStore finished", 0xFF210099)
End Sub

Public Sub DeleteVectorStore(vectorStoreId As String)
	LogColor("Sub OpenAILibrary DeleteVectorStore started", 0xFF210099)
	Private Job1 As HttpJob
	' Initialize the HTTP job
	Job1.Initialize("DeleteVectorStore", Me)

	Job1.Delete("https://api.openai.com/v1/vector_stores/" & vectorStoreId) ' Send a DELETE request to the specified URL

	' Set the necessary headers
	Job1.GetRequest.SetHeader("Content-Type", "application/json")
	Job1.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	Job1.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary DeleteVectorStore finished", 0xFF210099)
End Sub

Public Sub RetrieveVectorStore(vectorStoreId As String)
	LogColor("Sub OpenAILibrary RetrieveVectorStore started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("RetrieveVectorStore", Me)
	'https://api.openai.com/v1/vector_stores/vs_abc123
	job.PostString("https://api.openai.com/v1/vector_stores/" & vectorStoreId,"")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("Content-Type", "application/json")
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogColor("Sub OpenAILibrary RetrieveVectorStore finished", 0xFF210099)
End Sub

'FINE TUNING JOBS

Public Sub ListFineTuningJobs(limit As String, after As String, before As String)
	LogColor("Sub OpenAILibrary ListFineTuningJobs started", 0xFF210099)
    
	Dim job As HttpJob
	job.Initialize("ListFineTuningJobs", Me)
    
	Dim url As String = "https://api.openai.com/v1/fine_tuning/jobs"
	Dim parameters As String = ""

	' Add limit parameter if it's not blank
	If limit <> "" Then
		parameters = parameters & $"limit=${limit}"$
	End If

	' Add after parameter if it's not blank
	If after <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"after=${after}"$
	End If

	' Add before parameter if it's not blank
	If before <> "" Then
		If parameters <> "" Then parameters = parameters & "&"
		parameters = parameters & $"before=${before}"$
	End If

	' If parameters are present, append them to the URL
	If parameters <> "" Then
		url = url & "?" & parameters
	End If

	job.Download(url)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("ListFineTuningJobs", url)
	LogColor("Sub OpenAILibrary ListFineTuningJobs finished", 0xFF210099)
End Sub

Public Sub RetrieveFineTuningJob(fine_tuning_job_id As String)
	LogColor("Sub OpenAILibrary RetrieveFineTuningJob started", 0xFF210099)

	Dim job As HttpJob
	job.Initialize("RetrieveFineTuningJob", Me)

	' Construct URL
	Dim url As String = $"https://api.openai.com/v1/fine_tuning/jobs/${fine_tuning_job_id}"$

	job.Download(url)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("Content-Type", "application/json")

	' New lines added for support
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogHttpRequest("RetrieveFineTuningJob", url)
End Sub

Public Sub CancelFineTuningJob(fine_tuning_job_id As String)
	LogColor("Sub OpenAILibrary CancelFineTuningJob started", 0xFF210099)

	Dim job As HttpJob
	job.Initialize("CancelFineTuningJob", Me)

	' Construct URL for cancellation
	Dim url As String = $"https://api.openai.com/v1/fine_tuning/jobs/${fine_tuning_job_id}/cancel"$

	job.PostString(url, "")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("Content-Type", "application/json")

	' New lines added for support
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogHttpRequest("CancelFineTuningJob", url)
End Sub

'FINE TUNING EVENTS

Public Sub ListFineTuningEvents(FineTuningJobID As String, limit As String, after As String)
	LogColor("Sub OpenAILibrary ListFineTuningEvents started", 0xFF210099)
    
	' Start with the base URL
	Dim baseUrl As String = $"https://api.openai.com/v1/fine_tuning/jobs/${FineTuningJobID}/events"$

	Dim url As String = baseUrl
	Dim parameters As String = ""

	' Add the after parameter if it's not blank
	If after <> "" Then
		If parameters <> "" Then parameters = parameters & "&" ' Add '&' if parameters are already present
		parameters = parameters & $"after=${after}"$
	End If
    
	' Add the limit parameter if it's not blank
	If limit <> "" Then
		If parameters <> "" Then parameters = parameters & "&" ' Add '&' if parameters are already present
		parameters = parameters & $"limit=${limit}"$
	End If

	' Check if we have any parameters to append
	If parameters <> "" Then
		url = url & "?" & parameters ' Append parameters to the URL
	End If

	Dim job As HttpJob
	job.Initialize("ListFineTuningEvents", Me)
	job.Download(url)  ' Download the constructed URL
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
    
	' Log the request
	LogHttpRequest("ListFineTuningEvents", url)
	LogColor("Sub OpenAILibrary ListFineTuningEvents finished", 0xFF210099)
End Sub

'MODELs

Public Sub ListModels
	LogColor("Sub OpenAILibrary ListModels started", 0xFF210099)
	Dim job As HttpJob
	job.Initialize("ListModels", Me)
	job.Download("https://api.openai.com/v1/models")
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogHttpRequest("ListModels", "https://api.openai.com/v1/models") ' Log the request.
	LogColor("Sub OpenAILibrary ListModels finished", 0xFF210099)
End Sub

Public Sub DeleteFineTunedModel(ModelID As String)
	LogColor("Sub OpenAILibrary DeleteFineTunedModel started", 0xFF210099)

	Dim job As HttpJob
	job.Initialize("DeleteFineTunedModel", Me)

	' Construct URL
	Dim url As String = $"https://api.openai.com/v1/models/${ModelID}"$

	' Send the DELETE request
	job.Delete(url)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("Content-Type", "application/json")

	' New lines added for support
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogHttpRequest("DeleteFineTunedModel", url)
End Sub

Public Sub RetrieveModel(ModelID As String)
	LogColor("Sub OpenAILibrary RetrieveModel started", 0xFF210099)

	Dim job As HttpJob
	job.Initialize("RetrieveModel", Me)

	' Construct URL
	Dim url As String = $"https://api.openai.com/v1/models/${ModelID}"$

	job.Download(url)
	job.GetRequest.SetHeader("Authorization", "Bearer " & apiKey)
	job.GetRequest.SetHeader("Content-Type", "application/json")

	' New lines added for support
	job.GetRequest.SetHeader("OpenAI-Beta", "assistants=v2")
	LogHttpRequest("RetrieveModel", url)
End Sub

'JOBDONE

Private Sub JobDone(Job As HttpJob)
	LogColor("Sub OpenAILibrary JobDone started", 0xFF210099)
	Log("Job Done Sub Starting")
	Log("Job Name = " & Job.JobName)
	
	If Job.Success Then
		'Log("Job.GetString = " & Job.GetString)
		'Note that Job.Getstring will give a string with JSON format.
		
		'This Select...Case is to create a resultMap for the Cases that have a JSON response.
		Select Job.JobName
			Case "ttsJob", "DownloadImage"
				'No action needed here.  These two don't give a JSON response.
			Case Else
				Dim NewObject As String = Job.JobName
				' Create a Map from the JSON object
				Dim parser As JSONParser
				parser.Initialize(Job.GetString)
				' Assuming parser is already initialized and pointing to the JSON structure
				Dim resultMap As Map = parser.NextObject()
		End Select
		
		'This Select...Case is to route the response back to the main code.  Special cases first then Case Else for all the rest.
		Select Job.JobName
			
			Case "DownloadImage"
				Log("Here in Library Download Image Case.")
				'Log("Here DownloadImage1")
				Dim inputStream As InputStream = Job.GetInputStream
				If xui.SubExists(mTarget, mEventname & "_ImageResponse", 1) Then
					CallSub2(mTarget, mEventname & "_ImageResponse", inputStream)
				Else
					Log("OpenAI_ImageResponse(inputStream As InputStream) sub is missing from user code.")
				End If
				
			Case "generateimage"
				Dim url As String = resultMap.Get("data").As(List).Get(0).As(Map).Get("url")
				'Log("Here for logging - generate image case")
	
				Log("OpenAI: Image response ready!")
    
				' Check if the resultMap contains "data" and handle it accordingly
				Dim data As List
				data = resultMap.Get("data")
    
				' Check if we have at least one entry in data
				If data.Size = 0 Then
					Return ' Exit if the data array has no entries
				End If

				' Extract the first entry from the data array
				Dim imageData As Map = data.Get(0)
    
				' Extract required fields
				Dim revisedPrompt As String = imageData.Get("revised_prompt")
				Dim url As String = imageData.Get("url")
				Dim created As Long = resultMap.Get("created") ' Getting created timestamp from resultMap
				'Log("URL = " & url)
				' Construct a new response map
				Dim finalResponseMap As Map
				finalResponseMap.Initialize
				finalResponseMap.Put("object", "imageresponse") ' Adding the "object" level
				finalResponseMap.Put("revised_prompt", revisedPrompt)
				finalResponseMap.Put("url", url)
				finalResponseMap.Put("created", created)

				' Convert finalResponseMap to JSON string to pass to main code
				Dim jsonGen As JSONGenerator
				jsonGen.Initialize(finalResponseMap)
				Dim resultJson As String = jsonGen.ToString
				'LogUserAction("resultJson = " & resultJson)
				
				' Call back to the main code with the constructed JSON response
				If xui.SubExists(mTarget, mEventname & "_Response", 2) Then
					CallSub2(mTarget, mEventname & "_Response", resultJson)
				Else
					xui.MsgboxAsync("There is no Sub OpenAI_Response(response As String) in the user's code.", "ERROR")
				End If

				Dim j As HttpJob
				j.Initialize("DownloadImage", Me)
				'LogHttpRequest("DownloadImage", url) ' Log the request
				Log("Download URL = " & url)
				j.Download(url)
				
			Case "ttsJob"
				Dim audioStream As InputStream = Job.GetInputStream
				If xui.SubExists(mTarget, mEventname & "_TTSResponse", 1) Then
					CallSub2(mTarget, mEventname & "_TTSResponse", audioStream)
				Else
					Log("OpenAI_TTSResponse(inputStream As InputStream) sub is missing from user code.")
				End If

				'Those responses that require the object to be changed are handled here:
				'The reason for the object change is becaause many of the responses from OpenAI have the same object.  The main code needs a way to differentiate between
				'jobs to respond correctly.
			Case "ListModels", "DeleteFineTunedModel", "ListFineTuningJobs", "ListFineTuningEvents", "ListFineTuningJobs", "CheckRunStatus", "DeleteVectorStore", "RetrieveVectorStore", "ListVectorStores", "ListRuns", "RetrieveRun", "GetMessagesFromThread", "RetrieveAssistant", "ListAssistants", "listFilesJob", "CancelUpload", "completeUploadJob", "deleteFileJob", "UploadFile"

				' Modify the "object" field
				resultMap.Put("object", NewObject)
				
				' Reassemble the Map to JSON
				Dim jsonGen As JSONGenerator
				jsonGen.Initialize(resultMap)
    
				Dim newJson As String = jsonGen.ToString ' Create the new JSON string

				' Log the newly created JSON
				Log("Modified JSON: " & newJson)
				If xui.SubExists(mTarget, mEventname & "_Response", 2) Then
					CallSub2(mTarget, mEventname & "_Response",newJson)
				Else
					xui.MsgboxAsync("There is no Sub OpenAI_Response(response As String) in the user's code.","ERROR")
				End If
				
			Case Else 'Handles every job that is not a special case and does not need the Object changed.
				
				Log("CASE ELSE in JobDone")
				If xui.SubExists(mTarget, mEventname & "_Response", 2) Then
					CallSub2(mTarget, mEventname & "_Response",Job.GetString)
				Else
					xui.MsgboxAsync("There is no Sub OpenAI_Response(response As String) in the user's code.","ERROR")
				End If
		End Select
	Else
		' Call error sub for error handling
		HandleError(Job.JobName & ": " & Job.ErrorMessage)
	End If
	Job.Release
	LogColor("Sub OpenAILibrary JobDone finished", 0xFF210099)
End Sub

Private Sub HandleError(errorMessage As String)
	LogColor("Sub OpenAILibrary HandleError started", 0xFF210099)
	If xui.SubExists(mTarget, mEventname & "_Error", 1) Then
		CallSub2(mTarget, mEventname & "_Error", errorMessage)
	End If
	LogColor("Sub OpenAILibrary HandleError finished", 0xFF210099)
End Sub

Private Sub LogHttpRequest(jobName As String, request As String)
	'LogColor("Sub OpenAILibrary LogHttpRequest started", 0xFF210099)
	Log("HTTP Request Job Name: " & jobName)
	Log("HTTP Request String: " & request)
	'LogColor("Sub OpenAILibrary LogHttpRequest finished", 0xFF210099)
End Sub

